/* eslint camelcase: "off" */
const { execFile } = require('child_process');
const { accessSync, constants } = require('fs');
const FfmpegProgram = require('../FfmpegProgram/FfmpegProgram');
const FfmpegInput = require('../FfmpegInput/FfmpegInput');
const FfmpegSettings = require('../FfmpegSettings/FfmpegSettings');
const FfmpegMediaFile = require('../FfmpegMediaFile/FfmpegMediaFile');
const FfmpegStream = require('../FfmpegStream/FfmpegStream');
const { FfmpegUnsupportedError } = require('../FfmpegErrors/FfmpegErrors');
const { toInteger } = require('../FfmpegValidators/FfmpegValidators');
const { FfmpegFrameratePresets } = require('../FfmpegProgram/FfmpegProgram.helpers');

const { env: { DRAWTEXTFONT: DrawTextFont, FFMPEG_BIN: FfmpegBin } } = process;

const validTargetPrefixes = ['ntsc', 'pal', 'film', 'dummy'];
const validTargets = ['vcd', 'svcd', 'dvd', 'dv', 'dv50'];

function parseFilterPads(str) {
  const pads = [];
  const regex = /\[([^\]]+)\]/g;
  let matches = regex.exec(str);

  while (matches) {
    const [, pad] = matches;
    pads.push(pad);
    matches = regex.exec(str);
  }

  return pads;
}

function parseFilterGraphString(filtergraph) {
  const regex = /^(\[\S+\])*(\w+)(?:=(\w+=[^[]+))*(\[\S+\])*/g;

  const [, inputPads, filterName, filterSettings, outputPads] = regex.exec(filtergraph);
  const filterObj = { [filterName]: {} };

  if (inputPads) {
    filterObj[filterName].inputPads = parseFilterPads(inputPads);
  }

  if (outputPads) {
    filterObj[filterName].outputPads = parseFilterPads(outputPads);
  }

  if (filterSettings) {
    filterObj[filterName].settings = {};
    const settings = filterSettings.split(':');
    settings.forEach((setting) => {
      const [settingName, settingValue] = setting.split('=');
      filterObj[filterName].settings[settingName] = settingValue;
    });
  }

  return filterObj;
}

function reduceFilterObject(filters = []) {
  const filterGraphStrings = filters.map((filter) => {
    let filterGraphString = '';
    const filterName = Object.keys((filter))[0];
    const options = filter[filterName];

    if (options.inputPads) {
      const pads = options.inputPads.map(pad => `[${pad}]`);
      filterGraphString += pads.join('');
    }

    filterGraphString += filterName;

    if (options.settings) {
      const filterSettings = Object.keys(options.settings).map(key => `${key}=${options.settings[key]}`);
      filterGraphString += `=${filterSettings.join(':')}`;
    }

    if (options.outputPads) {
      const pads = options.outputPads.map(pad => `[${pad}]`);
      filterGraphString += pads.join('');
    }

    return filterGraphString;
  });

  return filterGraphStrings.join(',');
}

function formatTimecode(timecode) {
  let fields = timecode.split(/[.:;]/);
  const dropFrame = timecode.includes(';');
  const escapedColon = '\\\\:';
  const finalSeparator = dropFrame ? '\\;' : escapedColon;

  fields.forEach((field) => {
    if (field.length > 2) {
      throw new Error('');
    }
  });


  fields = fields.map(field => `${field}`.padStart(2, '0'));

  while (fields.length < 4) {
    fields.push('00');
  }

  return `${fields[0]}${escapedColon}${fields[1]}${escapedColon}${fields[2]}${finalSeparator}${fields[3]}`;
}

function addFilter(filterGraph, filterList, supportedFilters) {
  switch (typeof filterGraph) {
    case 'string': // e.g. 'scale=w=1280:h=720'
      addFilter(parseFilterGraphString(filterGraph), filterList, supportedFilters);
      break;
    case 'object':
      if (Array.isArray(filterGraph)) { // Array of filters
        filterGraph.forEach(filter => addFilter(filter, filterList, supportedFilters));
      } else if (Object.keys(filterGraph).length === 1) { // Object with one filter
        const filterName = Object.keys(filterGraph)[0];
        if (supportedFilters.includes(filterName)) {
          const copy = JSON.parse(JSON.stringify(filterGraph));
          filterList.push(copy);
        } else {
          throw new FfmpegUnsupportedError(`${filterName} is not a supported filter`);
        }
      } else { // Object with multiple filters
        Object.keys(filterGraph).forEach((key) => {
          addFilter({ [key]: filterGraph[key] }, filterList, supportedFilters);
        });
      }
      break;
    default:
      throw new TypeError('');
  }
}

class FfmpegOutput extends FfmpegMediaFile {
  constructor(inputFile) {
    if (!(inputFile instanceof FfmpegInput)) {
      throw new TypeError('new FfmpegOutput requires an FfmpegInput as its argument');
    }

    super();
    this.input = inputFile;
    this.setDefaultMap();

    this.ffmpegSettings = new FfmpegSettings();

    this.programs = [];
    this.videoFilters = [];
    this.audioFilters = [];
    this.complexFilters = [];
  }

  /**
   * Set the audio codec.
   * @param {string} codec
   */
  acodec(codec) {
    if (FfmpegProgram.getSupportedAudioEncoders().includes(codec)) {
      this.addOption('c:a', codec);
      return this;
    }

    throw new FfmpegUnsupportedError(`Audio codec "${codec}" is not supported on your version of Ffmpeg.`);
  }

  /**
   * Create the filtergraph specified by filtergraph and use it to filter the stream.
   *
   * @param filtergraph
   */
  af(filterGraph) {
    try {
      addFilter(filterGraph, this.audioFilters, FfmpegProgram.getSupportedAudioFilters());
    } catch (error) {
      if (error instanceof TypeError) {
        throw new TypeError('af() requires an object or filtergraph string argument');
      } else {
        throw error;
      }
    }

    return this;
  }

  /**
   * Set the number of audio frames to output.
   *
   * @param {number} number
   */
  aframes(number) {
    this.addOption('-frames:a', toInteger(number));
    return this;
  }

  /**
   * Set the audio quality (codec-specific, VBR).
   */
  aq(q) {
    if (q) {
      this.addOption('-q:a', q);
    }

    return this;
  }

  /**
  * Add an attachment to the output file. This is supported by a few formats like Matroska for
  * e.g. fonts used in rendering subtitles. Attachments are implemented as a specific type of
  * stream, so this option will add a new stream to the file. It is then possible to use
  * per-stream options on this stream in the usual way. Attachment streams created with this
  * option will be created after all the other streams (i.e. those created with -map or automatic
  * mappings).
  * @param {string} filename
  */
  attach(filename) {
    accessSync(filename, constants.R_OK);

    this.addOption('-attach', filename);
    return this;
  }

  /**
    * Set the number of data frames to output.
    */
  dframes(number) {
    this.addOption('-frames:d', toInteger(number));
    return this;
  }

  filter_complex(filterGraph) {
    try {
      addFilter(filterGraph, this.complexFilters, FfmpegProgram.getSupportedFilters());
    } catch (error) {
      if (error instanceof TypeError) {
        throw new TypeError('filter_complex requires an object or filtergraph string argument');
      } else {
        throw error;
      }
    }

    return this;
  }

  /**
   * Stop writing to the stream after framecount frames.
   * @param {number} framecount
   */
  frames(framecount) {
    this.addOption('-frames', toInteger(framecount));
    return this;
  }

  /**
   * Set the file size limit, expressed in bytes. No further chunk of bytes is written after the
   * limit is exceeded. The size of the output file is slightly more than the requested file size.
   * @param {number} limit_size
   */
  fs(limit_size) {
    this.addOption('-fs', toInteger(limit_size));
    return this;
  }

  getStream(stream) {
    if (stream instanceof FfmpegStream) {
      return stream;
    }

    if (/^\d+(?::(?:[avsdV]|\d+)(?::\d+)?)?$/.test(stream)) {
      const elems = stream.split(':');

      if (elems.length === 1) {
        return this.input.streams;
      }

      if (elems.length === 2) {
        switch (elems[1]) {
          case 'v':
            return this.input.videoStreams();
          case 'V':
            return this.input.largestDimensions();
          case 'a':
            return this.input.audioStreams();
          case 's':
            return this.input.subtitleStreams();
          case 'd':
            return this.input.dataStreams();
          default:
            return this.input.streams[parseInt(elems[1], 10)];
        }
      }

      if (elems.length === 3) {
        const streams = this.getStream(`${elems[0]}:${elems[1]}`);
        return streams[parseInt(elems[2], 10)];
      }
    }

    throw new TypeError(`Invalid map string ${stream}`);
  }

  /**
   * Designate one or more input streams as a source for the output file. Each input stream is
   * identified by the input file index input_file_id and the input stream index input_stream_id
   * within the input file. Both indices start at 0. If specified, sync_file_id:stream_specifier
   * sets which input stream is used as a presentation sync reference.
   *
   * @param {FfmpegStream[]}
   *
   * @returns this
   */
  map(stream, ...moreStreams) {
    // using this method will clear the default map, but not custom maps
    if (this.defaultMap === true) {
      this.streams = [];
      this.defaultMap = false;
    }

    if (Array.isArray(stream)) {
      stream.forEach(element => this.map(element));
    } else if (stream instanceof FfmpegStream) {
      this.streams.push(stream);
    } else if (stream === 0) {
      this.map(this.input.streams);
    } else if (typeof stream === 'string') {
      this.map(this.getStream(stream));
    } else {
      throw new TypeError('map() requires a string or an FfmpegStream, (usually from an FfmpegInput)');
    }

    if (moreStreams.length > 0) {
      moreStreams.forEach(element => this.map(element));
    }

    return this;
  }

  /**
   * Creates a program with the specified title, program_num and adds the specified stream(s) to it.
   *
   * @param {number|string} title
   * @param {number|string} program_num
   * @param {number|string} stream
   */
  program(streams, { title, program_num } = {}) {
    let selectedStreams;

    if (Array.isArray(streams)) {
      selectedStreams = streams.map(stream => this.getStream(stream));
    } else {
      selectedStreams = this.getStream(streams);
    }

    const selectedProgramNum = program_num === undefined ? null : program_num;
    const selectedTitle = title === undefined ? null : title;

    this.programs.push({
      title: selectedTitle,
      program_num: selectedProgramNum,
      streams: selectedStreams,
    });

    return this;
  }

  /**
   * Alias for {@link qscale}
   */
  q(q) {
    return this.qscale(q);
  }

  /**
   * Use fixed quality scale (VBR). The meaning of q/qscale is codec-dependent. If qscale() is
   * called on an output rather than a stream then it applies only to the video stream, this is to
   * maintain compatibility with previous behavior and as specifying the same codec specific value
   * to 2 different codecs that is audio and video generally is not what is intended when no
   * stream_specifier is used.
   * @param {number} q
   */
  qscale(q) {
    const asInteger = toInteger(q);
    this.addOption('-qscale', asInteger);
    return this;
  }

  /**
   * Specify target file type ("vcd", "svcd", "dvd", "dv", "dv50"). type may be prefixed with
   * "pal-", "ntsc-" or "film-" to use the corresponding standard. All the format options (bitrate,
   * codecs, buffer sizes) are then set automatically.
   * @param {string} type
   */
  target(type) {
    const elements = type.split('-');
    while (elements.length < 2) {
      elements.unshift('dummy');
    }

    if (validTargetPrefixes.includes(elements[0]) && validTargets.includes(elements[1])) {
      this.addOption('-target', type);
      return this;
    }

    throw new TypeError(`${type} is not a valid argument for target()`);
  }

  /**
   * Set the video codec.
   * @param {string} codec
   */
  vcodec(codec) {
    if (FfmpegProgram.getSupportedVideoEncoders().includes(codec)) {
      this.addOption('-c:v', codec);
      return this;
    }

    throw new FfmpegUnsupportedError(`Video codec "${codec}" is not supported on your version of Ffmpeg.`);
  }

  /**
   * Create the filtergraph specified by filtergraph and use it to filter the stream.
   *
   * @param filtergraph
   */
  vf(filterGraph) {
    try {
      addFilter(filterGraph, this.videoFilters, FfmpegProgram.getSupportedVideoFilters());
    } catch (error) {
      if (error instanceof TypeError) {
        throw new TypeError('vf() requires an object or filtergraph string argument');
      } else {
        throw error;
      }
    }

    return this;
  }

  /**
   * Set the number of video frames to output.
   */
  vframes(number) {
    const asInteger = toInteger(number);
    this.addOption('-frames:v', asInteger);
    return this;
  }

  /**
   * Force video tag/fourcc.
   */
  vtag(fourccOrTag) {
    if (!fourccOrTag) {
      return this;
    }

    this.addOption('-tag:v', fourccOrTag);
    return this;
  }

  /**
   * Set the audio sample format.
   *
   * @param {string}
   */
  sample_fmt(sample_fmt) {
    if (!sample_fmt) {
      return this;
    }

    if (!FfmpegProgram.getSupportedSampleFmts().includes(sample_fmt)) {
      throw new FfmpegUnsupportedError(`${sample_fmt} is not a supported sample_fmt`);
    }

    this.addOption('-sample_fmt', sample_fmt);
    return this;
  }

  /**
   * Sets this instance  to have the same stream mapping as Ffmpeg's defualts:
   * Largest video stream, audio stream with the most channels, and first subtitle stream
   */
  setDefaultMap() {
    this.streams = [];
    this.defaultMap = false;

    const video = this.input.largestDimensions();
    const audio = this.input.mostChannels();

    const allSubs = this.input.subtitleStreams();
    const subs = allSubs.length > 1 ? allSubs[0] : allSubs;

    if (video) this.map(video);
    if (audio) this.map(audio);
    if (subs) this.map(subs);

    this.defaultMap = true;
    return this;
  }

  /*
   * Add a scaling filter, keeping the aspect ratio, and given only one dimension.
  */
  resize({
    width,
    height,
    preset,
    stream = this.input.largestDimensions(this.input.streams),
    algorithm = 'fast_bilinear',
  }) {
    if (preset) {
      // eslint-disable-next-line prefer-destructuring
      width = FfmpegProgram.getResizePreset(preset).width; // eslint-disable-line no-param-reassign
      // eslint-disable-next-line prefer-destructuring
      height = FfmpegProgram.getResizePreset(preset).height;// eslint-disable-line no-param-reassign
    }

    if (width && !height) {
      // eslint-disable-next-line no-param-reassign
      height = Math.trunc((stream.height / stream.width) * width);
    }

    if (height && !width) {
      // eslint-disable-next-line no-param-reassign
      width = Math.trunc((stream.width / stream.height) * height);
    }

    return this.vf({ scale: { settings: { width, height, sws_flags: algorithm } } });
  }

  addTimecodeBurn(start, r, {
    fontfile = DrawTextFont,
    fontsize = '(h/16)',
    fontcolor = 'white',
    box = true,
    boxcolor = 'black',
    x = '((w-tw)-(w/10))',
    y = '(h/10)',
  } = {}) {
    if (!/\d+(?:(?:\/|\.)\d+)?/.test(r) && !Object.prototype.hasOwnProperty.call(FfmpegFrameratePresets, r)) {
      throw new Error(`Rate is a required parameter, and must be a decimal number, rational, or ffmpeg preset. (You passed ${r})`);
    }

    let timecode;
    try {
      timecode = formatTimecode(start);
    } catch (error) {
      throw new TypeError(`Timecode must be a string in the format hh:mm:ss:ff. (You passed ${start})`);
    }

    return this.vf({
      drawtext: {
        timecode,
        fontfile,
        fontsize,
        r,
        x,
        y,
        fontcolor,
        box,
        boxcolor,
      },
    });
  }

  getVideoFilterArguments() {
    if (this.videoFilters.length === 0) {
      return [];
    }

    return ['-filter:v', `${reduceFilterObject(this.videoFilters)}`];
  }

  getAudioFilterArguments() {
    if (this.audioFilters.length === 0) {
      return [];
    }

    return ['-filter:a', `${reduceFilterObject(this.audioFilters)}`];
  }

  getComplexFilterArguments() {
    if (this.complexFilters.length === 0) {
      return [];
    }

    return ['-filter_complex', `${reduceFilterObject(this.complexFilters)}`];
  }

  getMapArguments() {
    return this.streams.reduce((acc, stream) => acc.concat(['-map', `0:${stream.index}`]), []);
  }

  getProgramArguments() {
    const commandLine = [];

    this.programs.forEach((program) => {
      let programString = '';
      if (program.title) {
        programString += `title=${program.title}:`;
      }
      if (program.program_num) {
        programString += `program_num=${program.program_num}:`;
      }

      programString += program.streams.map(stream => `st=${this.streams.indexOf(stream)}`).join(':');

      commandLine.push('-program', programString);
    });

    return commandLine;
  }

  getCommandLine() {
    return [].concat(
      this.ffmpegSettings.getCommandLine(),
      this.getVideoFilterArguments(),
      this.getAudioFilterArguments(),
      this.getProgramArguments(),
      this.getMapArguments(),
    );
  }

  transcode(outputPath) {
    return execFile(
      FfmpegBin,
      [].concat(
        this.input.getCommandLine(),
        this.getCommandLine(),
        [outputPath],
      ),
    );
  }
}

module.exports = FfmpegOutput;
